@@ -1,20 +1,37 @@
 # -*- coding: utf-8 -*-
-# Copyright (C) 2004-2017 Aaron Swartz
+# Copyright (C) 2004-2021 Aaron Swartz
+#                         Amir Yalon <105565+amiryal@users.noreply.github.com>
+#                         Amir Yalon <git@please.nospammail.net>
+#                         Anders Damsgaard <anders@adamsgaard.dk>
 #                         Andrey Zelenchuk <azelenchuk@parallels.com>
+#                         Andrey Zelenchuk <azelenchuk@plesk.com>
 #                         Brian Lalor
 #                         Dean Jackson
 #                         Dennis Keitzel <github@pinshot.net>
+#                         Emil Oppeln-Bronikowski <emil@fuse.pl>
 #                         Erik Hetzner
 #                         Etienne Millon <me@emillon.org>
 #                         George Saunders <georgesaunders@gmail.com>
+#                         Gregory Soutade <gregory@soutade.fr>
 #                         J. Lewis Muir <jlmuir@imca-cat.org>
 #                         Jakub Wilk <jwilk@jwilk.net>
 #                         Joey Hess
+#                         Jonathan Kamens <jik@kamens.us>
+#                         Kaashif Hymabaccus <kaashif@kaashif.co.uk>
+#                         Karthikeyan Singaravelan <tir.karthi@gmail.com>
 #                         Lindsey Smith <lindsey.smith@gmail.com>
+#                         Léo Gaspard <leo@gaspard.io>
 #                         Marcel Ackermann
 #                         Martin 'Joey' Schulze
+#                         Martin Monperrus <monperrus@users.noreply.github.com>
+#                         Martin Vietz <git@martin.vietz.eu>
 #                         Matej Cepl
+#                         Notkea <pacien@users.noreply.github.com>
+#                         Profpatsch <mail@profpatsch.de>
 #                         W. Trevor King <wking@tremily.us>
+#                         auouymous <5005204+auouymous@users.noreply.github.com>
+#                         auouymous <au@qzx.com>
+#                         shan3141 <xpoh434@users.noreply.github.com>
 #
 # This file is part of rss2email.
 #
@@ -33,20 +50,26 @@
 """Define the ``Feed`` class for handling a single feed
 """
 
+import calendar as _calendar
 import collections as _collections
+import platform
+from email.message import Message
 from email.mime.message import MIMEMessage as _MIMEMessage
 from email.mime.multipart import MIMEMultipart as _MIMEMultipart
 from email.utils import formataddr as _formataddr
+from email.utils import formatdate as _formatdate
+from email.utils import parseaddr as _parseaddr
 import hashlib as _hashlib
 import html.parser as _html_parser
 import re as _re
 import socket as _socket
 import time as _time
-import urllib.error as _urllib_error
 import urllib.request as _urllib_request
 import uuid as _uuid
 import xml.sax as _sax
 import xml.sax.saxutils as _saxutils
+from functools import lru_cache
+from typing import Optional, Dict, Any, Tuple
 
 import feedparser as _feedparser
 import html2text as _html2text
@@ -61,8 +84,6 @@ from . import error as _error
 from . import util as _util
 
 
-_USER_AGENT = 'rss2email/{} ({})'.format(__version__, __url__)
-_feedparser.USER_AGENT = _USER_AGENT
 _urllib_request.install_opener(_urllib_request.build_opener())
 _SOCKET_ERRORS = []
 for e in ['error', 'herror', 'gaierror']:
@@ -142,7 +163,7 @@ class Feed (object):
     >>> test_section = CONFIG.pop('feed.test-feed')
 
     """
-    _name_regexp = _re.compile('^[\w\d.-]+$')
+    _name_regexp = _re.compile(r'^[\w\d.-]+$')
 
     # saved/loaded from feed.dat using __getstate__/__setstate__.
     _dynamic_attributes = [
@@ -163,6 +184,9 @@ class Feed (object):
     _default_configured_attributes[
         _default_configured_attributes.index('from')
         ] = 'from_email'  # `from` is a Python keyword
+    _default_configured_attributes[
+        _default_configured_attributes.index('user_agent')
+        ] = '_user_agent'  # `user_agent` is a getter that does substitution
     # all attributes that are saved/loaded from .config
     _configured_attributes = (
         _non_default_configured_attributes + _default_configured_attributes)
@@ -185,6 +209,7 @@ class Feed (object):
         'date_header',
         'trust_guid',
         'trust_link',
+        'reply_changes',
         'html_mail',
         'use_css',
         'unicode_snob',
@@ -208,6 +233,12 @@ class Feed (object):
         'digest_post_process',
         ]
 
+    @property
+    def user_agent(self):
+        return self._user_agent.\
+            replace('__VERSION__', __version__).\
+            replace('__URL__', __url__)
+
     def __init__(self, name=None, url=None, to=None, config=None):
         self._set_name(name=name)
         self.reset()
@@ -215,6 +246,7 @@ class Feed (object):
                 (attr, getattr(self, attr))
                 for attr in self._dynamic_attributes))
         self.load_from_config(config=config)
+        self._fix_user_agent() # Fix feeds broken by user agent change in 3.11
         if url:
             self.url = url
         if to:
@@ -319,7 +351,7 @@ class Feed (object):
         """
         self.etag = None
         self.modified = None
-        self.seen = {}
+        self.seen = {} # type: Dict[str, Dict[str, Any]]
 
     def _set_name(self, name):
         if not self._name_regexp.match(name):
@@ -348,8 +380,10 @@ class Feed (object):
         timeout = config.getint('feed-timeout')
         kwargs = {}
         if proxy:
-            kwargs['handlers'] = [_urllib_request.ProxyHandler({'http':proxy})]
-        f = _util.TimeLimitedFunction(timeout, _feedparser.parse)
+            kwargs['handlers'] = [
+                _urllib_request.ProxyHandler({ 'http': proxy, 'https': proxy })
+            ]
+        f = _util.TimeLimitedFunction('feed {}'.format(self.name), timeout, _feedparser.parse)
         return f(self.url, self.etag, modified=self.modified, **kwargs)
 
     def _process(self, parsed):
@@ -359,14 +393,14 @@ class Feed (object):
             _LOG.debug('processing {}'.format(entry.get('id', 'no-id')))
             processed = self._process_entry(parsed=parsed, entry=entry)
             if processed:
-                guid,id_,sender,message = processed
+                guid, _, sender, message = processed
                 if self.post_process:
                     message = self.post_process(
                         feed=self, parsed=parsed, entry=entry, guid=guid,
                         message=message)
                     if not message:
                         continue
-                yield (guid, id_, sender, message)
+                yield processed
 
     def _check_for_errors(self, parsed):
         warned = False
@@ -378,9 +412,11 @@ class Feed (object):
             self.url = parsed['url']
         elif status not in [200, 302, 304, 307]:
             raise _error.HTTPError(status=status, feed=self)
+
         http_headers = parsed.get('headers', {})
         if http_headers:
             _LOG.debug('HTTP headers: {}'.format(http_headers))
+            http_headers = dict((k.lower(), v) for k, v in http_headers.items())
         if not http_headers:
             _LOG.warning('could not get HTTP headers: {}'.format(self))
             warned = True
@@ -396,7 +432,7 @@ class Feed (object):
         if version:
             _LOG.debug('feed version {}'.format(version))
         else:
-            _LOG.warning('unrecognized version: {}'.format(self))
+            _LOG.debug('unrecognized version: {}'.format(self))
             warned = True
 
         exc = parsed.get('bozo_exception', None)
@@ -449,30 +485,43 @@ class Feed (object):
                 return default
             raise
 
-    def _process_entry(self, parsed, entry):
-        id_ = self._get_entry_id(entry)
-        # If .trust_guid isn't set, we get back hashes of the content.
-        # Instead of letting these run wild, we put them in context
-        # by associating them with the actual ID (if it exists).
-        guid = entry.get('id', id_)
-        if isinstance(guid, dict):
-            guid = guid.values()[0]
-        # In some bad RSS feeds, id is present but empty...
-        if guid == '':
-            guid = id_
-        if guid in self.seen:
-            if self.seen[guid]['id'] == id_:
-                _LOG.debug('already seen {}'.format(id_))
-                return  # already seen
-        _LOG.debug('not seen {}'.format(id_))
+    def _process_entry(self, parsed, entry) -> Optional[Tuple[str, Dict[str, Any], str, Message]]:
+        guid = self._get_uid_for_entry(entry)
+        new_hash = self._get_entry_hash(entry)
+
+        old_state = self.seen.get(guid)
+        if old_state is None:
+            _LOG.debug('not seen {}'.format(guid))
+            new_state = {} # type: Dict[str, Any]
+        else:
+            _LOG.debug('already seen {}'.format(guid))
+            if 'old' in old_state:
+                del old_state['old']
+            if self.reply_changes:
+                if new_hash != old_state.get('hash'):
+                    _LOG.debug('hash changed for {}'.format(guid))
+                    new_state = old_state.copy()
+                else:
+                    return None
+            else:
+                return None
+
+        new_state['hash'] = new_hash
+
         sender = self._get_entry_email(parsed=parsed, entry=entry)
         subject = self._get_entry_title(entry)
+
+        message_id = '<{0}@{1}>'.format(_uuid.uuid4(), platform.node())
+        in_reply_to = old_state.get('message_id') if old_state is not None else None
         extra_headers = _collections.OrderedDict((
                 ('Date', self._get_entry_date(entry)),
-                ('Message-ID', '<{}@dev.null.invalid>'.format(_uuid.uuid4())),
-                ('User-Agent', _USER_AGENT),
+                ('Message-ID', message_id),
+                ('In-Reply-To', in_reply_to),
+                ('User-Agent', self.user_agent),
+                ('List-ID', '<{}.localhost>'.format(self.name)),
+                ('List-Post', 'NO (posting not allowed on this list)'),
                 ('X-RSS-Feed', self.url),
-                ('X-RSS-ID', id_),
+                ('X-RSS-ID', guid),
                 ('X-RSS-URL', self._get_entry_link(entry)),
                 ('X-RSS-TAGS', self._get_entry_tags(entry)),
                 ))
@@ -506,31 +555,43 @@ class Feed (object):
             extra_headers=extra_headers,
             config=self.config,
             section=self.section)
-        return (guid, id_, sender, message)
 
-    def _get_entry_id(self, entry):
-        """Get best ID from an entry."""
+        return guid, new_state, sender, message
+
+    def _get_uid_for_entry(self, entry) -> str:
+        """Get the best UID (unique ID) for the entry."""
         if self.trust_link:
-            return entry.get('link', None)
+            uid = self._get_entry_link(entry)
+            if uid:
+                return uid
         if self.trust_guid:
-            if getattr(entry, 'id', None):
-                # Newer versions of feedparser could return a dictionary
-                if isinstance(entry.id, dict):
-                    return entry.id.values()[0]
-                return entry.id
+            uid = self._get_entry_id(entry)
+            if uid:
+                return uid
+        return self._get_entry_hash(entry)
+
+    @lru_cache(1)
+    def _get_entry_hash(self, entry) -> str:
         content = self._get_entry_content(entry)
         content_value = content['value'].strip()
         if content_value:
-            return _hashlib.sha1(
-                content_value.encode('unicode-escape')).hexdigest()
+            text = content_value
         elif getattr(entry, 'link', None):
-            return _hashlib.sha1(
-                entry.link.encode('unicode-escape')).hexdigest()
+            text = entry.link
         elif getattr(entry, 'title', None):
-            return _hashlib.sha1(
-                entry.title.encode('unicode-escape')).hexdigest()
+            text = entry.title
+        else:
+            text = ""
+        return _hashlib.sha1(text.encode('unicode-escape')).hexdigest()
 
-    def _get_entry_link(self, entry):
+    def _get_entry_id(self, entry) -> Optional[str]:
+        id_ = getattr(entry, 'id', None)
+        # Newer versions of feedparser could return a dictionary
+        if isinstance(id_, dict):
+            return next(iter(id_.values()), None)
+        return id_
+
+    def _get_entry_link(self, entry) -> Optional[str]:
         return entry.get('link', None)
 
     def _get_entry_title(self, entry):
@@ -555,7 +616,7 @@ class Feed (object):
                 if entry.get(kind, None):
                     datetime = entry[kind]
                     break
-        return _time.strftime("%a, %d %b %Y %H:%M:%S -0000", datetime)
+        return _formatdate(_calendar.timegm(datetime))
 
     def _get_entry_name(self, parsed, entry):
         """Get the best name
@@ -705,6 +766,7 @@ class Feed (object):
         if taglist:
             return ','.join(taglist)
 
+    @lru_cache(1)
     def _get_entry_content(self, entry):
         """Select the best content from an entry.
 
@@ -725,9 +787,9 @@ class Feed (object):
         if entry.get('summary_detail', None):
             contents.append(entry.summary_detail)
         if self.html_mail:
-            types = ['text/html', 'text/plain']
+            types = ['application/xhtml+xml', 'text/html', 'text/plain']
         else:
-            types = ['text/plain', 'text/html']
+            types = ['text/plain', 'text/html', 'application/xhtml+xml']
         for content_type in types:
             for content in contents:
                 if content['type'] == content_type:
@@ -748,15 +810,16 @@ class Feed (object):
             if self.use_css and self.css:
                 lines.extend([
                         '    <style type="text/css">',
-                        self.css,
+                        _saxutils.escape(self.css),
                         '    </style>',
                         ])
             lines.extend([
                     '</head>',
-                    '<body>',
+                    '<body dir="auto">',
                     '<div id="entry">',
                     '<h1 class="header"><a href="{}">{}</a></h1>'.format(
-                        link, subject),
+                        _saxutils.escape(link) if link else '',
+                        _saxutils.escape(subject)),
                     '<div id="body">',
                     ])
             if content['type'] in ('text/html', 'application/xhtml+xml'):
@@ -764,27 +827,28 @@ class Feed (object):
             else:
                 lines.append(_saxutils.escape(content['value'].strip()))
             lines.append('</div>')
-            lines.extend([
-                    '<div class="footer">'
-                    '<p>URL: <a href="{0}">{0}</a></p>'.format(link),
-                    ])
+            lines.append('<div class="footer">')
+            if link:
+                lines.append(
+                    '<p>URL: <a href="{0}">{0}</a></p>'.format(
+                        _saxutils.escape(link)))
             for enclosure in getattr(entry, 'enclosures', []):
                 if getattr(enclosure, 'url', None):
                     lines.append(
                         '<p>Enclosure: <a href="{0}">{0}</a></p>'.format(
-                            enclosure.url))
+                            _saxutils.escape(enclosure.url)))
                 if getattr(enclosure, 'src', None):
                     lines.append(
                         '<p>Enclosure: <a href="{0}">{0}</a></p>'.format(
-                            enclosure.src))
+                            _saxutils.escape(enclosure.src)))
                     lines.append(
-                        '<p><img src="{}" /></p>'.format(enclosure.src))
+                        '<p><img src="{}" /></p>'.format(_saxutils.escape(enclosure.src)))
             for elink in getattr(entry, 'links', []):
                 if elink.get('rel', None) == 'via':
                     url = elink['href']
                     title = elink.get('title', url)
                     lines.append('<p>Via <a href="{}">{}</a></p>'.format(
-                            url, title))
+                            _saxutils.escape(url), _saxutils.escape(title)))
             lines.extend([
                     '</div>',  # /footer
                     '</div>',  # /entry
@@ -823,10 +887,10 @@ class Feed (object):
         section = self.section
         if section not in self.config:
             section = 'DEFAULT'
-        _email.send(sender=sender, recipient=self.to, message=message,
+        _email.send(recipient=self.to, message=message,
                     config=self.config, section=section)
 
-    def run(self, send=True):
+    def run(self, send=True, clean=False):
         """Fetch and process the feed, mailing entry emails.
 
         >>> feed = Feed(
@@ -841,42 +905,80 @@ class Feed (object):
         """
         if not self.to:
             raise _error.NoToEmailAddress(feed=self)
+        if clean:
+            self.etag = None
+            self.modified = None
         parsed = self._fetch()
 
+        if clean and len(parsed.entries) > 0:
+            for guid in self.seen:
+                self.seen[guid]['old'] = True
+
         if self.digest:
             digest = self._new_digest()
             seen = []
-
-        for (guid, id_, sender, message) in self._process(parsed):
-            _LOG.debug('new message: {}'.format(message['Subject']))
-            if self.digest:
-                seen.append((guid, id_))
+            for (guid, state, sender, message) in self._process(parsed):
+                _LOG.debug('new message: {}'.format(message['Subject']))
+                seen.append((guid, state))
                 self._append_to_digest(digest=digest, message=message)
-            else:
+            if seen:
+                if self.digest_post_process:
+                    digest = self.digest_post_process(feed=self, parsed=parsed, seen=seen, message=digest)
+                    if not digest:
+                        return
+                _LOG.debug('new digest for {}'.format(self))
+                if send:
+                    self._send_digest(digest=digest, sender=sender)
+                for (guid, state) in seen:
+                    self.seen[guid] = state
+        else:
+            for (guid, state, sender, message) in self._process(parsed):
+                _LOG.debug('new message: {}'.format(message['Subject']))
                 if send:
                     self._send(sender=sender, message=message)
-                if guid not in self.seen:
-                    self.seen[guid] = {}
-                self.seen[guid]['id'] = id_
-
-        if self.digest and seen:
-            if self.digest_post_process:
-                digest = self.digest_post_process(
-                    feed=self, parsed=parsed, seen=seen, message=digest)
-                if not digest:
-                    return
-            self._send_digest(
-                digest=digest, seen=seen, sender=sender, send=send)
+                    state['message_id'] = str(message["Message-ID"])
+                self.seen[guid] = state
 
         self.etag = parsed.get('etag', None)
         self.modified = parsed.get('modified', None)
 
+        if clean and len(parsed.entries) > 0:
+            # A feed might only show the newest N entries, but if the feed
+            # author deletes a new entry, an older entry will reappear. Deleting
+            # all entries not in the feed could cause an old entry to be resent.
+            # To avoid this, the three newest entries no longer in the feed are
+            # kept, allowing up to three new entries to be deleted from the feed
+            # before anything would be resent.
+
+            # Entries from new feeds are added oldest to newest, and new entries
+            # are always inserted at the end. Entries no longer in the feed will
+            # have an 'old' key added above. This iterates over the entry keys
+            # (guid) from last to first, ignoring entries still in the feed. The
+            # three newest 'old' entries are skipped and all others are deleting.
+
+            # Python 3.7 guarantees Dict insertion order and CPython has done so
+            # since 3.5. It is unlikely older and other implementations would
+            # have stored the entries out-of-order, but if they did, the three
+            # entries kept won't be correct. And in the unlikely event the feed
+            # author deletes an entry, an old entry could be resent.
+
+            old = 3
+            for guid in reversed(list(self.seen)):
+                if 'old' in self.seen[guid]:
+                    if old > 0:
+                        del self.seen[guid]['old']
+                    else:
+                        del self.seen[guid]
+                    old = old - 1
+
     def _new_digest(self):
         digest = _MIMEMultipart('digest')
-        digest['To'] = self.to  # TODO: _Header(), _formataddr((recipient_name, recipient_addr))
+        digest['To'] = _formataddr(_parseaddr(self.to))  # Encodes with utf-8 as necessary
         digest['Subject'] = 'digest for {}'.format(self.name)
-        digest['Message-ID'] = '<{}@dev.null.invalid>'.format(_uuid.uuid4())
-        digest['User-Agent'] = _USER_AGENT
+        digest['Message-ID'] = '<{0}@{1}>'.format(_uuid.uuid4(), platform.node())
+        digest['User-Agent'] = self.user_agent
+        digest['List-ID'] = '<{}.localhost>'.format(self.name)
+        digest['List-Post'] = 'NO (posting not allowed on this list)'
         digest['X-RSS-Feed'] = self.url
         return digest
 
@@ -885,22 +987,33 @@ class Feed (object):
         part.add_header('Content-Disposition', 'attachment')
         digest.attach(part)
 
-    def _send_digest(self, digest, seen, sender, send=True):
+    def _send_digest(self, digest, sender):
         """Send a digest message
 
         The date is extracted from the last message in the digest
         payload.  We assume that this part exists.  If you don't have
         any messages in the digest, don't call this function.
         """
-        digest['From'] = sender  # TODO: _Header(), _formataddr()...
+        digest['From'] = sender
         last_part = digest.get_payload()[-1]
         last_message = last_part.get_payload()[0]
         digest['Date'] = last_message['Date']
+        self._send(sender=sender, message=digest)
 
-        _LOG.debug('new digest for {}'.format(self))
-        if send:
-            self._send(sender=sender, message=digest)
-        for (guid, id_) in seen:
-            if guid not in self.seen:
-                self.seen[guid] = {}
-            self.seen[guid]['id'] = id_
+    def _fix_user_agent(self):
+        """Fix feed user agent settings broken in 3.11
+
+        Release 3.11 added code to allow the user agent string to be
+        configured globally and on a per-feed basis. It was supposed
+        to allow __VERSION__ and __URL__ in configuration setting to
+        be substituted dynamically with the current version and URL of
+        rss2email, but a bug in the implementation caused them to be
+        substituted statically in the settings for any newly added
+        feed. We've fixed that problem, but we want to go back now and
+        repair feeds that got the wrong user agent value in the
+        interim.
+        """
+        if self._user_agent == \
+           'rss2email/3.11 (https://github.com/rss2email/rss2email)':
+            self._user_agent = 'rss2email/__VERSION__ (__URL__)'
+            self.save_to_config()
