@@ -1,10 +1,13 @@
-# Copyright (C) 2012-2020 Andrey Zelenchuk <azelenchuk@parallels.com>
+# Copyright (C) 2012-2021 Andrey Zelenchuk <azelenchuk@parallels.com>
 #                         Andrey Zelenchuk <azelenchuk@plesk.com>
 #                         Gregory Soutade <gregory@soutade.fr>
 #                         Kaashif Hymabaccus <kaashif@kaashif.co.uk>
+#                         Lucas <lucas@sexy.is>
 #                         Léo Gaspard <leo@gaspard.io>
 #                         Profpatsch <mail@profpatsch.de>
 #                         W. Trevor King <wking@tremily.us>
+#                         auouymous <5005204+auouymous@users.noreply.github.com>
+#                         ryneeverett <ryneeverett@gmail.com>
 #
 # This file is part of rss2email.
 #
@@ -103,6 +106,9 @@ def run(*args, **kwargs):
         '-n', '--no-send', dest='send',
         default=True, action='store_const', const=False,
         help="fetch feeds, but don't send email")
+    run_parser.add_argument(
+        '--clean', action='store_true',
+        help='clean old feed entries')
     run_parser.add_argument(
         'index', nargs='*',
         help='feeds to fetch (defaults to fetching all feeds)')
@@ -169,7 +175,12 @@ def run(*args, **kwargs):
     # Immediately lock so only one r2e instance runs at a time
     if UNIX:
         import fcntl as _fcntl
-        lockfile_path = _os.path.join(_os.environ.get("XDG_RUNTIME_DIR", default="/tmp/"), "rss2email.lock")
+        from pathlib import Path as _Path
+        dir = _os.environ.get("XDG_RUNTIME_DIR")
+        if dir is None:
+            dir = _os.path.join("/tmp", "rss2email-{}".format(_os.getuid()))
+            _Path(dir).mkdir(mode=0o700, parents=True, exist_ok=True)
+        lockfile_path = _os.path.join(dir, "rss2email.lock")
         lockfile = open(lockfile_path, "w")
         _fcntl.lockf(lockfile, _fcntl.LOCK_EX)
         _LOG.debug("acquired lock file {}".format(lockfile_path))
@@ -183,6 +194,8 @@ def run(*args, **kwargs):
         feeds = _feeds.Feeds(datafile_path=args.data, configfiles=args.config)
         if args.func != _command.new:
             feeds.load()
+        if not args.verbose:
+            _LOG.setLevel(feeds.config['DEFAULT']['verbose'].upper())
         args.func(feeds=feeds, args=args)
     except _error.RSS2EmailError as e:
         e.log()
